# Copyright 2009-2011 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

""" General oops request lifetime stuff, abstracted from WSGI things

"""

__all__ = ["OOPSReporter", "DEFAULT_SERIALIZER_FACTORY"]

import sys
import logging

from storm.tracer import install_tracer, remove_tracer_type

from canonical.oops import serializer
from canonical.oops.oops import OOPS, OOPSLog, OOPSLoggingHandler
from canonical.oops.stormtracer import OOPSStorm, OOPSStormTracer

DEFAULT_SERIALIZER_FACTORY = 'canonical.oops.serializer.OOPSBz2Serializer'


class OOPSReporter(object):
   """An object encapsulating OOPS reporting and oops allocation."""

   def __init__(self, oops_dir,
                serializer_factory_name=DEFAULT_SERIALIZER_FACTORY,
                debug=False, key="appserver", thread_local=True):
        """Install tracers and set up the oops serializer."""
        self.thread_local = thread_local
        # Add the oops to the storm tracing
        remove_tracer_type(OOPSStormTracer)
        install_tracer(OOPSStormTracer())
        # Add the oops to the standard logging
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            if isinstance(handler, OOPSLoggingHandler):
                break
        else:
            root_logger.addHandler(OOPSLoggingHandler())
        fh = debug and sys.stderr or None
        self.serializer = serializer.create(serializer_factory_name,
                                            key, oops_dir, fh=fh)

   def open_request(self):
       """Create a new OOPS context manager for a single request."""
       return OOPSRequest(self.serializer, thread_local=self.thread_local)


class OOPSRequest(object):
    """A context manager embodying the life cycle of a request."""

    def __init__(self, serializer, thread_local):
        """Create and retain an OOPS instance."""
        self.serializer = serializer
        self.oops = OOPS(self.serializer.get_next_oops_id)
        self.oops.add(OOPSStorm(thread_local=thread_local))
        self.oops.add(OOPSLog(thread_local=thread_local))

    def __enter__(self):
        """Enter the context, returning the context's OOPS instance."""
        return self.oops

    def __exit__(self, ex_type, ex_value, ex_tb):
        """Leave the context, reporting the OOPS if appropriate."""
        root_logger = logging.getLogger()

        if ex_type is not None:
            self.oops.keep()
            message = "Unhandled application exception " \
                      "(OOPS %s)" % (self.oops.id,)
            exc_info = (ex_type, ex_value, ex_tb)
            root_logger.error(message, exc_info=exc_info)

        if self.oops.wanted:
            try:
                self.serializer.dump(self.oops)
            except:
                exc_info = sys.exc_info()
                try:
                    message = "Error when dumping OOPS %s" % self.oops.id
                    root_logger.error(message, exc_info=exc_info)
                finally:
                    exc_info = None

        return False
