# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

""" test for canonical.oops.wsgi """
import os
import unittest

from canonical.oops.tests import testcase
from canonical.oops import wsgi
from paste.fixture import TestApp
from canonical.oops.wsgi import track_iterator_completion


class TestSimpleWSGIOops(testcase.WSGIOopsTestCase):
    """ simple test of wsgi oops just test for all ok and error """

    def test_simple_error(self):
        """ test an error from the application and check if
        there is an oopsid assigned
        """
        res = self.testapp.get('?error=forced_error',
                    status=500,
                    expect_errors=True)
        self.assertNotEquals(None, self.oops.id)
        self.assertTrue(self.oops.wanted)

    def test_all_ok(self):
        """ test the all ok path """
        res = self.testapp.get('?ok=all_ok', status=200, expect_errors=False)
        self.assertNotEquals(None, self.oops.id)
        self.assertFalse(self.oops.wanted)

    def test_hide_meta_false(self):
        """hide_meta is False by default so our query string param will be
        captured"""
        res = self.testapp.get('?error=forced_error',
                    status=500,
                    expect_errors=True)
        self.assertNotEquals(None, self.oops.id)
        self.assertTrue(self.oops.wanted)
        self.assertTrue('metadata' in self.oops.get_dump_data().keys())
        meta = self.oops.get_dump_data()['metadata']
        self.assertEquals('error=forced_error', meta['QUERY_STRING'])

    def test_hide_meta_true(self):
        """with hide_meta set to True our query string param will not be
        captured"""
        self.testapp.app.hide_meta = True
        res = self.testapp.get('?error=forced_error',
                    status=500,
                    expect_errors=True)
        self.assertNotEquals(None, self.oops.id)
        self.assertTrue(self.oops.wanted)
        self.assertFalse('metadata' in self.oops.get_dump_data().keys())

    def test_dump_oops_and_want_oops(self):
        """An oops serialization should occur for either dump_oops or want_oops
        variables being set"""
        # no dump but wanted +1
        res = self.testapp.get('?error=forced_error',
                    status=500,
                    expect_errors=True)
        inner_dir = os.listdir(self.tmp_dir)[0]
        reports = os.listdir(os.path.join(self.tmp_dir, inner_dir))
        self.assertEqual(1, len(reports))

        # dump but not wanted +1
        res = self.testapp.get('?ok=all_ok',
                    status=200,
                    expect_errors=False,
                               extra_environ={'dump-oops' : 'whatever'})
        reports = os.listdir(os.path.join(self.tmp_dir, inner_dir))
        self.assertEqual(2, len(reports))

        # neither dump nor wanted +0
        res = self.testapp.get('?ok=all_ok',
                    status=200,
                    expect_errors=False)
        reports = os.listdir(os.path.join(self.tmp_dir, inner_dir))
        self.assertEqual(2, len(reports))

def track_completion(body, on_completion):
    """Prove that we've been called by adding to the body."""
    yield "OK"
    iterator = track_iterator_completion(body, on_completion)
    for chunk in iterator:
        yield chunk


class TestResultIteration(testcase.WSGIOopsTestCase):
    """Test iteration on results."""

    def test_immediate_start(self):
        """Test that we start without waiting for body consumption."""

        def dummy_app(environ, start_response):
            """Produce a brief response."""
            start_response("200 OK", {}, exc_info=None)
            return ["foobar"]

        started = [False]

        def dummy_start_response(status, headers, exc_info=None):
            """Sets flag when response is started"""
            started[0] = True

        app = wsgi.OopsWare(dummy_app, self.tmp_dir)
        app({}, dummy_start_response)
        self.assertTrue(started[0])

    def test_result_tracking(self):
        """Test that our own result tracker gets called."""

        def dummy_app(environ, start_response):
            """Produce a simple response."""
            start_response("200 OK", [('Content-Type', 'text/plain')],
                           exc_info=None)
            return ["foobar"]

        def dummy_start_response(status, headers, exc_info=None):
            """Do nothing."""

        name = "canonical.oops.tests.test_wsgi.track_completion"
        app = wsgi.OopsWare(dummy_app, self.tmp_dir,
                            completion_tracker_name=name)
        body = app({}, dummy_start_response)
        result = list(body)
        self.assertEqual(result, ["OK", "foobar"])

    def _generate_garbage(self):
        """Generate an infinite sequence of garbage until closed."""
        while True:
            yield "xyzzy"

    def test_close_no_oops(self):
        """Test that closing the generator doesn't oops"""

        def dummy_app(environ, start_response):
            """Produce a response with infinite garbage."""
            start_response("200 OK", {}, exc_info=None)
            return self._generate_garbage()

        def dummy_start_response(status, headers, exc_info=None):
            """Do nothing."""

        interceptor = testcase.OopsWareInterceptor(dummy_app)
        app = wsgi.OopsWare(interceptor, self.tmp_dir)
        gen = app({}, dummy_start_response)
        gen.close()
        self.assertFalse(interceptor.oops.wanted)

    def test_close_closes_inner_generator(self):
        """Test that closing the outer generator closes the inner one."""
        garbage = self._generate_garbage()

        def dummy_app(environ, start_response):
            """Produce a response with the specific garbage generator."""
            start_response("200 OK", {}, exc_info=None)
            return garbage

        def dummy_start_response(status, headers, exc_info=None):
            """Do nothing."""

        interceptor = testcase.OopsWareInterceptor(dummy_app)
        app = wsgi.OopsWare(interceptor, self.tmp_dir)
        gen = app({}, dummy_start_response)
        gen.close()
        self.assertRaises(StopIteration, garbage.next)


class TestEnvironWSGIOops(testcase.WSGIOopsTestCase):
    """Create the oops enviroment for testing"""
    def create_test_app(self):
        """Create a useful test app for the wsgi testing."""
        def test_app(environ, start_response):
            """ simple wsgi app that returns a 500 error
            if can't find OOPSID in environ
            """
            if environ.get('OOPSID', None) is None:
                start_response("500 Missing OOPSID in environ",
                            [('Content-Type', 'text/html'),])
            else:
                start_response("200 OK", [('Content-Type', 'text/html'), ])
            return ['Avant!']
        return test_app

    def test_oopsid_in_environ(self):
        """ test that OOPSID is in the environ """
        res = self.testapp.get('?ok=all_ok', status=200, expect_errors=False)


class TestWSGIOopsKey(testcase.WSGIOopsTestCase):
    """ Check that the OOPSSerializer key can be configured when instantiating
        an OopsWare.
    """

    serializer_key = 'fooberry'

    def setUp(self):
        super(testcase.WSGIOopsTestCase, self).setUp()
        self.test_root_dir = None
        self.tmp_dir = self._make_tmp_dir()
        self.oops_ware_interceptor = testcase.OopsWareInterceptor(
            self.create_test_app())
        self.app = wsgi.OopsWare(self.oops_ware_interceptor, self.tmp_dir,
            key=self.serializer_key)
        self.testapp = TestApp(self.app)

    def test_oopsid_server_key(self):
        """ Test an error from the application and check that the serializer
            key and oops id are correctly assigned.
        """
        res = self.testapp.get('?error=forced_error',
                    status=500,
                    expect_errors=True)
        self.assert_(self.serializer_key in self.oops.id)


class TestWSGIOoops404(testcase.WSGIOopsTestCase):
    """Tests for collection of OOPS reports on 404 errors."""

    def make_testapp(self, oops_on_404=True):
        """Make a TestApp fixture using a specially configured OopsWare."""
        app = wsgi.OopsWare(self.oops_ware_interceptor, self.tmp_dir,
                            oops_on_404=oops_on_404)
        testapp = TestApp(app)
        return testapp

    def test_oops_generated_on_404_with_referer(self):
        """If a 404 is generated with a referer, then the report is kept."""
        testapp = self.make_testapp()
        res = testapp.get("?notfound", status=404, expect_errors=True,
                          headers={"Referer": "http://www.example.org/"})
        self.assertTrue(self.oops.wanted)

    def test_no_oops_generated_on_404_with_no_referer(self):
        """If a 404 is generated without a referer, the report is not kept."""
        testapp = self.make_testapp()
        res = testapp.get("?notfound", status=404, expect_errors=True)
        self.assertFalse(self.oops.wanted)

    def test_no_oops_if_disabled(self):
        """If generation of 404 reports is disabled, then the report
        is not kept."""
        testapp = self.make_testapp(oops_on_404=False)
        res = testapp.get("?notfound", status=404, expect_errors=True,
                          headers={"Referer": "http://www.example.org/"})
        self.assertFalse(self.oops.wanted)

