# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

""" WSGI Middleware Wrapper for oops

Use like so::

 app = OopsWare(app, oops_dir='/tmp/oops/')

oops_dir is a directory that will be created to dump oops in to.

"""

__metaclass__ = type
__all__ = ["OopsWare", "track_iterator_completion"]

import sys
import time
import logging

from canonical.oops.oops import OOPSMetaData
from canonical.oops.utils import resolve_name
from canonical.oops.reporter import OOPSReporter, DEFAULT_SERIALIZER_FACTORY

DEFAULT_COMPLETION_TRACKER = 'canonical.oops.wsgi.track_iterator_completion'

def track_iterator_completion(body, on_completion):
    """Wrap the given body iterator in one that tracks completion."""

    def iterate():
        """Track iterator completion."""
        try:
            yield None # throwaway value for priming
            for chunk in body:
                yield chunk
        except GeneratorExit:
            close = getattr(body, 'close', lambda: None)
            close()
            on_completion(None, None, None)
        except:
            on_completion(*sys.exc_info())
        else:
            on_completion(None, None, None)

    iterator = iterate()
    iterator.next() # prime generator
    return iterator


class OopsWare(object):
    """
    Intercept __call__ and act accordingly, returning the
    required page unless there was an error.
    """
    def __init__(self, app, oops_dir,
                 serializer_factory_name=DEFAULT_SERIALIZER_FACTORY,
                 completion_tracker_name=DEFAULT_COMPLETION_TRACKER,
                 debug=False, key="appserver", hide_meta=False,
                 oops_on_404=True):
        """
        Initialise the middleware for oops.
        """
        self.app = app
        self.hide_meta = hide_meta
        self.oops_on_404 = oops_on_404
        self.reporter = OOPSReporter(oops_dir=oops_dir,
                                     serializer_factory_name=
                                     serializer_factory_name,
                                     debug=debug, key=key)
        self.track_completion = resolve_name(completion_tracker_name)

    def __call__(self, environ, start_response):
        """
        Capture exceptions during WSGI application execution and dump
        oopses as needed.
        """
        root_logger = logging.getLogger()

        # Request start, record the time and create oops tracking object
        context = self.reporter.open_request()
        oops = context.__enter__()

        metadata = OOPSMetaData(environ)
        if not self.hide_meta:
            oops.add(metadata)

        environ['OOPSID'] = oops.assign_id()

        def trap_response(status, response_headers, exc_info=None):
            """
            Wrapper for start_response which traps the response code and
            related exception, if any.
            """
            status_code = status[0:3]

            metadata["http-code"] = status_code

            if status_code == '500':
                oops.keep()

            # The Referer check prevents random 404ed requests from logging
            if (self.oops_on_404 and status_code == '404' and
                environ.get('HTTP_REFERER', None)):
                oops.keep()

            try:
                if oops.wanted:
                    if not exc_info or not exc_info[0]:
                        exc_info = sys.exc_info()

                    if exc_info and exc_info[0]:
                        message = "WSGI application exception (OOPS %s)" % \
                                  oops.id
                        root_logger.error(message, exc_info=exc_info)

                return start_response(status, response_headers, exc_info)
            finally:
                exc_info = None

        setattr(trap_response, 'oops', oops)

        metadata["request-start"] = time.time()

        def on_completion(ex_type, ex_value, ex_tb):
            """Handle request completion."""
            metadata["request-end"] = time.time()
            if environ.has_key('dump-oops'):
                oops.keep()
            context.__exit__(ex_type, ex_value, ex_tb)

        try:
            body = self.app(environ, trap_response)
        except Exception:
            on_completion(*sys.exc_info())
        else:
            return self.track_completion(body, on_completion)
